import Vue from 'vue';
import * as _ from 'lodash';
import mockData from '@/datasets/harry.dataset.js';
import { sleep } from '@/mixins/helpers.js';
import { updateTags } from '@/store/modules/data/tags.js';
import {
  isMultilink,
  insertLinkWithMultilinkUpdate,
  getLinkSource,
  getLinkTarget,
  isNodeAggregated,
  isNodeBundle,
  isBundledIn,
  isReplacedBy,
  validateDataset,
  filterVisibleNodes,
  findConnectedLinks,
  arrGetNewId,
  findNodeById,
  filterVisibleLinks,
  isInMultilinkOfId,
  pushNodeToState,
  setVisAttr,
  arraySetVisAttrs,
  isLinkVisible,
  setVisAttrs,
  isInMultilink,
  removeVisAttr,
  makeNodesMap,
  findReplacementNode
} from './helpers';
import * as tags from './tags.js';

/**
 * Type definitions for node and link data structures
 *
 * @typedef { import('./typedefs').Node } Node
 * @typedef { import('./typedefs').NodeVis } NodeVis
 * @typedef { import('./typedefs').Link } Link
 * @typedef { import('./typedefs').LinkVis } LinkVis
 * @typedef { import('./tags').Tag } Tag
 */

export const namespaced = true;

export const state = {
  /** @type {Node[]} */
  nodes: [],
  /** @type {Link[]} */
  links: [],
  /** @type {Tag[]} */
  ...tags.state
};

export const mutations = {
  ...tags.getters,

  INIT(state, data) {
    state.nodes = _.cloneDeep(data.nodes);
    state.links = _.cloneDeep(data.links);
  },

  /**
   * Add new node
   * @param {Object} state
   * @param {Node} node
   */
  ADD_NODE(state, node) {
    node.id = arrGetNewId(state.nodes);
    state.nodes.push(node);
  },

  /**
   * Batch replace node attributes
   * @param {Object} state
   * @param {Object[]} payload const;
   * every diff will have node id and attributes to replace
   * [...{ id: nodeId, ...updateAttr: newVal}]
   * @param {Number} payload[].id
   *
   * @throws {Error} if diff array is empty or undefined
   * @throws {Error} if node is not found in state. Beware that previous valid entries will be commited
   */
  UPDATE_NODES(state, payload) {
    const nodesDiff = payload;
    if (!nodesDiff || nodesDiff.length < 1) throw Error('Update array undefined or empty');

    nodesDiff.forEach(ndiff => {
      if (!ndiff.id === undefined) throw Error('Entry has no id');
      const node = state.nodes.find(sn => sn.id === ndiff.id);
      if (!node) throw Error('Node not found');

      Vue.set(state.nodes, state.nodes.indexOf(node), _.merge(node, ndiff));
    });
  },

  /**
   * Aggregate nodes in general group, alias or bundle
   * Push new group node/links to state, update nodes/links.
   *
   * Aggregation types
   * ALIAS
   * a regular node (represents one entity, not a group);
   * merge links, attributes, merge conflict should be solved by user
   *
   * BUNDLING
   * hides a set of nodes in a selected node;
   * nodes are hidden, links are transferred (meaning they connect to the bundle node instead)
   *
   * GENERAL GROUP
   * nodes are hidden, links are transferred
   *
   * @param {Object}   state
   * @param {Object}   payload
   * @param {'group'|'alias'|'bundle'} payload.action aggregation type
   * @param {Number[]} payload.nodeIds to aggregate
   * @param {Node}     [payload.targetGroupNode] ref. to the new group node without ID, which will be set
   * @param {Number[]} [payload.removedLinkIds] OUT: will contain links hidden during aggregation, ordered from smallest
   * @param {Object[]} [payload.newVisibleLinks] OUT: will contain new links, ordered by id from smallest
   *
   * @throws {Error} when nodeIds are not given or less than 2
   * @throws {Error} when targetGroupNode already has an ID or is just regular node
   */
  AGGREGATE_NODES(state, payload) {
    const action = payload.action,
      nodesToGroupIds = payload.nodeIds,
      groupNode = payload.targetGroupNode,
      removedLinkIds = [],
      newVisibleLinks = [];

    if (action !== 'bundle' && (!nodesToGroupIds || nodesToGroupIds.length < 2))
      throw Error('Grouping requires at least 2 selected nodes');
    if (!groupNode) throw Error('Group node is undefined');
    if ((action === 'group' || action === 'alias') && groupNode.type === 'regular')
      throw Error('Node is regular.');
    if ((action === 'group' || action === 'alias') && groupNode.id !== undefined)
      throw Error('Group node should not come with an ID');

    const groupId = action === 'bundle' ? groupNode.id : pushNodeToState(state, groupNode, false);
    if (action === 'bundle') {
      const n = findNodeById(state, groupId);
      if (n.vis && n.vis.bundledNodes !== undefined) {
        const newBundledNodes = n.vis.bundledNodes;
        nodesToGroupIds.forEach(id => newBundledNodes.push(id));
        setVisAttr(n, 'bundledNodes', newBundledNodes);
      } else {
        setVisAttr(n, 'bundledNodes', nodesToGroupIds);
      }
      arraySetVisAttrs(state.nodes.filter(n => nodesToGroupIds.includes(n.id)), {
        bundledIn: groupId
      });
    } else {
      arraySetVisAttrs(state.nodes.filter(n => nodesToGroupIds.includes(n.id)), {
        replacedBy: groupId
      });
    }

    const updatedLinks = [];
    const newLinks = [];
    let currentNewLinkId = arrGetNewId(state.links);
    for (const link of state.links) {
      const isGroupTarget =
          nodesToGroupIds.includes(link.target) || (action === 'bundle' && groupId === link.target)
            ? true
            : false,
        isGroupSource =
          nodesToGroupIds.includes(link.source) || (action === 'bundle' && groupId === link.source)
            ? true
            : false,
        isGroupNewTarget =
          link.vis &&
          link.vis.newTarget !== undefined &&
          (nodesToGroupIds.includes(link.vis.newTarget) ||
            (action === 'bundle' && groupId === link.vis.newTarget))
            ? true
            : false,
        isGroupNewSource =
          link.vis &&
          link.vis.newSource !== undefined &&
          (nodesToGroupIds.includes(link.vis.newSource) ||
            (action === 'bundle' && groupId === link.vis.newSource))
            ? true
            : false,
        isSourceAndTargetSame =
          link.vis &&
          link.vis.newTarget !== undefined &&
          link.vis.newSource !== undefined &&
          link.vis.newSource === link.vis.newTarget
            ? true
            : false;

      let target = groupId,
        source = groupId;
      // ignore links not connected to subgraph
      if (!isGroupTarget && !isGroupSource && !isGroupNewTarget && !isGroupNewSource) {
        continue;
      }

      // remove all visible subgraph links from graph
      if (isLinkVisible(link)) removedLinkIds.push(link.id);

      // link was already transferred -> replace newSource, newTarget
      if (isGroupNewTarget || isGroupNewSource) {
        if (isSourceAndTargetSame) continue;
        if (isGroupNewTarget && !isGroupNewSource) source = link.vis.newSource;
        if (!isGroupNewTarget && isGroupNewSource) target = link.vis.newTarget;
        // link is new -> set newSource, newTarget for the first time
      } else if (!isMultilink(link) && (isGroupTarget || isGroupSource)) {
        if (isGroupTarget && !isGroupSource) source = link.source;
        if (!isGroupTarget && isGroupSource) target = link.target;
      } else if (isMultilink(link)) {
        link.vis.isArchived = true;
        continue;
      }

      // update link connection
      if (link.source !== source || link.target !== target) {
        setVisAttrs(link, { newSource: source, newTarget: target });
        if (
          action === 'bundle' &&
          (nodesToGroupIds.includes(link.source) || nodesToGroupIds.includes(link.target))
        ) {
          setVisAttr(link, 'forceTransferred', true);
        }
      }
      if (isInMultilink(link)) {
        Vue.delete(link.vis, 'inMultilink');
      }
      // ignore if inside the group
      if (source === target) {
        continue;
      }

      // check if updated links merge into multilinks
      const ml = insertLinkWithMultilinkUpdate(link, updatedLinks, groupId, currentNewLinkId);
      if (ml) {
        newLinks.push(ml);
        currentNewLinkId += 1;
      }
    }

    const updatedLinkIds = [];
    updatedLinks.forEach(l => {
      updatedLinkIds.push(l.id);
    });
    state.links
      .filter(l => updatedLinkIds.includes(l.id))
      .forEach(l => {
        const newer = updatedLinks.find(ul => ul.id === l.id);
        if (!newer) throw Error('Cannot find the new link');
        if (isInMultilink(newer)) {
          setVisAttrs(l, { inMultilink: newer.vis.inMultilink });
        }
      });

    newLinks.forEach(l => state.links.push(_.cloneDeep(l)));

    // updated not-in-multilink links and new multilinks will be shown
    updatedLinks
      .filter(l => !isInMultilink(l))
      .forEach(l => {
        if (isMultilink(l)) {
          newVisibleLinks.push({ id: l.id, source: getLinkSource(l), target: getLinkTarget(l) });
        } else {
          newVisibleLinks.push({ id: l.id, source: getLinkSource(l), target: getLinkTarget(l) });
        }
      });
    // set payload
    payload.removedLinkIds = _.sortBy(removedLinkIds, o => o);
    payload.newVisibleLinks = _.sortBy(newVisibleLinks, ['id']);

    updateTags(
      state.tags,
      state.nodes,
      nodesToGroupIds,
      action,
      groupNode,
      state.tagData ? state.tagData.color : ''
    );
  },

  /**
   * Reverse node aggregation for group, alias or bundle
   *
   * New data for graph will be placed in payload object.
   *
   * @param {Object}   state
   * @param {Object}   payload
   * @param {Number}   payload.nodeId group node to ungroup
   * @param {'group'|'bundle'} payload.revertAction
   * @param {boolean}  [payload.removeNode] OUT: will contain instruction whenever to remove the ungroped node
   * @param {Number[]} [payload.removedLinkIds] OUT: will contain removed links, ordered from smallest
   * @param {Object[]} [payload.newVisibleLinks] OUT: will contain replacement links for graph, ordered by ids from smallest
   * @param {Object[]} [payload.newVisibleNodes] OUT: will contain replacement nodes for graph, ordered by ids from smallest
   *
   * @throws {Error} target node does not exist
   * @throws {Error} target is regular node or not a bundle (in case of reverting bundling)
   * @throws {Error} target is a group and also bundles nodes
   */
  DISAGGREGATE_NODE(state, payload) {
    const revertAction = payload.revertAction,
      groupId = payload.nodeId,
      removedVisibleLinkIds = [],
      newVisibleLinks = [],
      newVisibleNodes = [];

    if (revertAction === undefined) throw Error('Undefined dissagregation action.');
    if (typeof groupId !== 'number' || groupId < 0) throw Error('Invalid argument.');
    const groupNode = state.nodes.find(n => n.id === groupId);
    if (!groupNode) throw Error('Target node was not found.');
    if (revertAction === 'group') {
      if (!isNodeAggregated(groupNode)) throw Error('Cannot revert grouping on regular node.');
      if (isNodeBundle(groupNode))
        throw Error('Cannot ungroup a bundled node. Revert bundle first.');
    } else if (revertAction === 'bundle' && !isNodeBundle(groupNode)) {
      throw Error('Cannot unbundle a node without bundled nodes.');
    }
    const removeTargetNode = revertAction !== 'bundle';

    // reveal original nodes
    const revealedNodeIds = [];
    if (revertAction === 'bundle') {
      // all types of target node
      const revealedNodes = state.nodes.filter(n => isBundledIn(n, groupId));
      revealedNodes.forEach(n => {
        revealedNodeIds.push(n.id);
        removeVisAttr(n, 'bundledIn');
        newVisibleNodes.push(
          n.vis && n.vis.graph
            ? { id: n.id, vis: { graph: _.cloneDeep(n.vis.graph) } }
            : { id: n.id }
        );
      });
      removeVisAttr(groupNode, 'bundledNodes');
    } else {
      const revealedNodes = state.nodes.filter(n => isReplacedBy(n, groupId));
      revealedNodes.forEach(n => {
        revealedNodeIds.push(n.id);
        removeVisAttr(n, 'replacedBy');
        newVisibleNodes.push(
          n.vis && n.vis.graph
            ? { id: n.id, vis: { graph: _.cloneDeep(n.vis.graph) } }
            : { id: n.id }
        );
      });
      // remove the target node
      Vue.set(state, 'nodes', state.nodes.filter(n => n.id !== groupId));
    }

    // convert state.nodes to hash map, because we will be doing recursive
    // lookups to update the connections of affected links
    const nodesMap = makeNodesMap(state);

    // divide links to two categories:
    //   - links untouched by ungrouping algorithm (for efficiency, as we do not need to iterate over them)
    //   - links that form the subgraph and do not originate from group
    // at the end, we will be merging them and replacing state.links
    // this is efficient, we are only creating vectors of references
    // however, we may want to deep copy these, if we will work with state.links
    const unaffectedLinks = [];
    const affectedLinks = [];
    const affectedMultilinks = [];
    const removedLinkIds = []; // will be used to calculate id for new multilinks, as we have split arrays and cannot know the available id
    for (const l of state.links) {
      // remove links originating from target node, when target is being removed
      // these also account for current multilinks, as they cannot be transferred
      if ((removeTargetNode || isMultilink(l)) && (l.source === groupId || l.target === groupId)) {
        removedLinkIds.push(l.id);
        if (isLinkVisible(l)) removedVisibleLinkIds.push(l.id);
        continue;
      }

      // find links transferred to group
      if (l.vis) {
        if (!isMultilink(l) && (getLinkSource(l) === groupId || getLinkTarget(l) === groupId)) {
          // normal links from subgraph
          affectedLinks.push(l);
          continue;
        } else if (l.vis.newSource === groupId || l.vis.newTarget === groupId) {
          affectedLinks.push(l);
          continue;
        } else if (
          isMultilink(l) &&
          (revealedNodeIds.includes(l.source) || revealedNodeIds.includes(l.target))
        ) {
          // multilinks from subgraph
          affectedMultilinks.push(l);
          continue;
        }
      }

      unaffectedLinks.push(l);
    }

    // iterate over subgraph links (those originating from group were already filtered out)
    //   - link is visible: remove from graph
    //   - is multilink, now what???
    //   - calculate newSource, newTarget
    //   - check against all these links, if we can create or recover a multilink
    const updatedAffectedLinks = [];
    const linksForIdCalculation = state.links.filter(l => !removedLinkIds.includes(l.id));

    // firstly recover old multilinks, so we then know, if we can just assign the link to them or create new multilinks
    for (const l of affectedMultilinks) {
      if (!isMultilink(l)) throw Error('Expected multilink');
      if (isLinkVisible(l)) removedVisibleLinkIds.push(l.id);
      updatedAffectedLinks.push(_.cloneDeep(l));
    }

    for (const l of affectedLinks) {
      // for simplicity, we remove all these links from graph and draw new after ungrouping, after multilinks formed
      if (isLinkVisible(l)) removedVisibleLinkIds.push(l.id);

      const updatedSource = findReplacementNode(nodesMap, l.source);
      const updatedTarget = findReplacementNode(nodesMap, l.target);

      removeVisAttr(l, 'inMultilink');
      if (revertAction === 'bundle') removeVisAttr(l, 'forceTransferred');

      // check, if we still transfer link or it is in original state
      if (l.source === updatedSource && l.target === updatedTarget) {
        removeVisAttr(l, 'newSource');
        removeVisAttr(l, 'newTarget');
      } else {
        l.vis.newSource = updatedSource;
        l.vis.newTarget = updatedTarget;
      }

      const ml = insertLinkWithMultilinkUpdate(
        l,
        updatedAffectedLinks,
        updatedSource,
        arrGetNewId(linksForIdCalculation)
      );
      if (ml) linksForIdCalculation.push(ml);
    }

    updatedAffectedLinks.forEach(l => {
      if (isLinkVisible(l))
        newVisibleLinks.push({
          id: l.id,
          source: l.vis && l.vis.newSource ? l.vis.newSource : l.source,
          target: l.vis && l.vis.newTarget ? l.vis.newTarget : l.target
        });
    });

    // now merge both link arrays and update state.links
    Vue.set(state, 'links', _.sortBy([...unaffectedLinks, ...updatedAffectedLinks], ['id']));
    // set payload
    payload.removedLinkIds = _.sortBy(removedVisibleLinkIds, o => o);
    payload.newVisibleLinks = _.sortBy(newVisibleLinks, ['id']);
    payload.newVisibleNodes = _.sortBy(newVisibleNodes, ['id']);
    payload.removeNode = removeTargetNode;
  }
};

export const actions = {
  ...tags.actions,

  fetchData({ commit }) {
    validateDataset(mockData);
    commit('INIT', mockData);
  },

  simulateBatchGrouping({ commit }) {
    let nonGroupedNodeIds = [];
    filterVisibleNodes(state.nodes).forEach(n => nonGroupedNodeIds.push(n.id));
    const groupedNodesIds = [];

    const updateGraph = async () => {
      // grouping
      while (nonGroupedNodeIds.length >= 2) {
        await sleep(300);
        let chooseCount = _.random(2, (state.nodes.length / 6) >> 0);
        if (chooseCount > nonGroupedNodeIds.length) chooseCount = nonGroupedNodeIds.length;
        if (chooseCount < 2) chooseCount = 2;

        let chosenIds = [];
        while (chosenIds.length < chooseCount) {
          const chosenId = nonGroupedNodeIds[_.random(0, nonGroupedNodeIds.length - 1)];
          chosenIds.push(chosenId);
          nonGroupedNodeIds = nonGroupedNodeIds.filter(n => n !== chosenId);
        }
        if (chosenIds.length < 2) throw Error('chosen less then 2 nodes for grouping');

        const chosenType = _.random(0, 100) < 80 ? 'group' : 'alias';
        const groupNode = { type: chosenType, label: chosenType === 'group' ? 'Group' : 'Alias' };
        commit('AGGREGATE_NODES', {
          action: chosenType,
          nodeIds: chosenIds,
          targetGroupNode: groupNode
        });

        chosenIds.forEach(id => groupedNodesIds.push(id));
      }
    };
    updateGraph();
  },

  import({ commit }, data) {
    validateDataset(data);
    commit('INIT', data);
  },

  createNode({ commit }, node) {
    commit('ADD_NODE', node);
    validateDataset(state);
  },

  updateNodes({ commit }, nodesDiff) {
    commit('UPDATE_NODES', nodesDiff);
  },

  /**
   * @param {Object} _
   * @param {Object} pl
   * @param {number[]} pl.nodes to group
   * @param {string} pl.label to group
   */
  groupNodes({ commit, dispatch }, pl) {
    let groupNode = {
      type: 'group',
      name: `Group of '${pl.label}'`,
      label: pl.label
    };

    commit('AGGREGATE_NODES', {
      action: 'group',
      nodeIds: pl.nodes,
      targetGroupNode: groupNode
    });
    validateDataset(state);
    dispatch('interaction/respondToRemovedNodes', pl.nodes, { root: true });
  },

  /**
   * @param {Object} _
   * @param {Object} pl
   * @param {number[]} pl.nodes to group
   * @param {string} pl.label to group
   */
  aliasNodes({ commit, dispatch }, pl) {
    let groupNode = {
      name: `Alias of '${pl.label}'`,
      type: 'alias',
      label: pl.label
    };

    commit('AGGREGATE_NODES', {
      action: 'alias',
      nodeIds: pl.nodes,
      targetGroupNode: groupNode
    });
    validateDataset(state);
    dispatch('interaction/respondToRemovedNodes', pl.nodes, { root: true });
  },

  /**
   * @param {object} vuexContext
   * @param {object} payload
   * @param {number} payload.representativeNodeId the "bundle" node
   * @param {number[]} payload.nodeIds to bundle inside the representative node
   */
  bundleNodes({ commit, getters, dispatch }, payload) {
    console.log('bundleNodes: ' + JSON.stringify(payload));
    commit('AGGREGATE_NODES', {
      action: 'bundle',
      nodeIds: payload.nodeIds,
      targetGroupNode: getters.getNodeById(payload.representativeNodeId)
    });
    validateDataset(state);
    dispatch('interaction/respondToRemovedNodes', payload.nodeIds, { root: true });
  },

  ungroupNode({ commit, getters, dispatch }, nodeId) {
    // unbundle first if bundle
    if (getters.getNodeBundledNodeCount(nodeId) > 0) {
      commit('DISAGGREGATE_NODE', {
        revertAction: 'bundle',
        nodeId: nodeId
      });
    }

    commit('DISAGGREGATE_NODE', {
      revertAction: 'group',
      nodeId: nodeId
    });
    validateDataset(state);
    dispatch('interaction/respondToRemovedNodes', [nodeId], { root: true });
  },

  unbundleNode({ commit }, nodeId) {
    commit('DISAGGREGATE_NODE', {
      revertAction: 'bundle',
      nodeId: nodeId
    });
    validateDataset(state);
  }
};

export const getters = {
  ...tags.getters,

  getState: state => state,
  getNodes: state => state.nodes,
  getLinks: state => state.links,

  getVisibleNodes: state => filterVisibleNodes(state.nodes),
  getVisibleLinks: state => filterVisibleLinks(state.links),

  getNodeById: state => id => findNodeById(state, id),
  getNodesByIds: state => ids => state.nodes.filter(n => ids.includes(n.id)),
  getLinkById: state => id => state.links.find(l => l.id === id),
  getMultilinkLinks: state => multilinkId =>
    state.links.filter(l => isInMultilinkOfId(l, multilinkId)),

  getLinkIdsConnectedToNode: state => nodeId => findConnectedLinks(state.links, nodeId),
  getNodeBundledNodeCount: state => id => {
    const n = findNodeById(state, id);
    if (!n) return 0;
    return n.vis && n.vis.bundledNodes !== undefined ? n.vis.bundledNodes.length : 0;
  },
  getNodesInAggregation: state => parentId => state.nodes.filter(n => isReplacedBy(n, parentId)),
  getNodesInBundle: state => bundleId => state.nodes.filter(n => isBundledIn(n, bundleId))
};
