import Vue from 'vue';
import * as _ from 'lodash';

/**
 * Type definitions for node and link data structures
 *
 * @typedef { import('./typedefs').Node } Node
 * @typedef { import('./typedefs').NodeVis } NodeVis
 * @typedef { import('./typedefs').Link } Link
 * @typedef { import('./typedefs').LinkVis } LinkVis
 */

/**
 * Get node from state
 * @param {Object} state
 * @param {Node[]} state.nodes
 * @param {number} id of node
 *
 * @returns {Node} Node instance, or undefined if not found
 */
export function findNodeById(state, id) {
  return state.nodes.find(n => n.id === id);
}

export function filterVisibleLinks(links) {
  return links.filter(l => isLinkVisible(l));
}

/**
 * Tell if node is visible in graph visualization
 * @param {Node} node
 *
 * @returns {boolean}
 */
export function isNodeVisible(node) {
  return !node.vis || (node.vis.replacedBy === undefined && node.vis.bundledIn === undefined);
}

export function filterVisibleNodes(nodes) {
  return nodes.filter(n => isNodeVisible(n));
}

export function isNodeAggregated(n) {
  return ['group', 'alias'].includes(n.type);
}

export function isNodeBundle(node) {
  return (
    node.vis &&
    node.vis.bundledNodes !== undefined &&
    Array.isArray(node.vis.bundledNodes) &&
    node.vis.bundledNodes.length > 0
  );
}

export function findConnectedLinks(links, nodeId) {
  const connected = links.filter(
    l => !isInMultilink(l) && (getLinkSource(l) === nodeId || getLinkTarget(l) === nodeId)
  );
  const sorted = _.sortBy(connected, [l => !isMultilink(l), l => l.id]); // multilinks first, then ids
  return sorted.map(l => l.id);
}

/**
 * Check if visible: not in multilink, not an archived multilink, newTarget != newSource
 * @param {Link} l
 *
 * @returns {boolean}
 */
export function isLinkVisible(l) {
  if (isInMultilink(l)) return false;
  if (l.vis) {
    if (l.vis.isMultilink && l.vis.isArchived) return false;
    if (l.vis.newSource !== undefined && l.vis.newSource === l.vis.newTarget) return false;
  } else {
    if (l.source === l.target) return false;
  }
  return true;
}

/**
 * Find first available id by iterating over objects in array
 * @param {Object[]} arr of objects, each object has attribute id
 *
 * @returns {number} id
 */
export function arrGetNewId(arr) {
  if (arr === undefined || !Array.isArray(arr)) throw 'Calling arrGetNewId with undefined array';
  if (arr.length === 0) return 0;

  return arr.reduce((prev, curr) => (prev.id > curr.id ? prev : curr)).id + 1;
}

function isLinkSimilar(link, linkToMatch) {
  const source = getLinkSource(linkToMatch);
  const target = getLinkTarget(linkToMatch);

  if (link.vis && link.vis.newSource !== undefined && link.vis.newTarget !== undefined) {
    return (
      (source === link.vis.newSource && target === link.vis.newTarget) ||
      (target === link.vis.newSource && source === link.vis.newTarget)
    );
  }

  return (
    (source === link.source && target === link.target) ||
    (target === link.source && source === link.target)
  );
}

/**
 * Return link.vis.newSource uif set, otherwise just link.source
 * @param {Link} link
 *
 * @return {number} ID of source
 */
export function getLinkSource(link) {
  if (link.vis && link.vis.newSource !== undefined) return link.vis.newSource;
  return link.source;
}

/**
 * Return link.vis.newTarget if set, otherwise just link.target
 * @param {Link} link
 *
 * @return {number} ID of target
 */
export function getLinkTarget(link) {
  if (link.vis && link.vis.newTarget !== undefined) return link.vis.newTarget;
  return link.target;
}

/**
 * @summary Insert a link in given array, checking for links with similar endpoints, updating the array.
 * If similar link exists, assign the link to a multilink in the array and update the multilink direction.
 * Recover archived multilink in array if possible. Create new multilink if necessary.
 *
 * To enable recovery of archived multilinks, add them into the array before calling this.
 *
 * @param {Link} link to insert; NOT MUTATED
 * @param {Link[]} arr to be updated with the new link
 * @param {number} groupId in order to set new multilink's source always to the group node
 * @param {number} availableId for creating new multilink
 *
 * @returns {Link} multilink, if new one was created during insertion; otherwise null
 */
export function insertLinkWithMultilinkUpdate(link, arr, groupId, availableId) {
  if (isMultilink(link))
    throw Error(
      'Bad usage of insertLinkWithMultilinkUpdate. Inserting a multilink is not allowed.'
    );

  const similarLink = arr.find(l => isLinkSimilar(l, link));
  if (!similarLink) {
    arr.push(_.cloneDeep(link));
    return null;
  }

  const source = getLinkSource(link);
  const target = getLinkTarget(link);

  if (isMultilink(similarLink)) {
    Vue.delete(similarLink.vis, 'isArchived');
    setVisAttr(link, 'inMultilink', similarLink.id);
    arr.push(_.cloneDeep(link));
    multilinkUpdateDirection(arr, similarLink);
  } else if (isInMultilink(similarLink)) {
    const ml = arr.find(l => l.id === similarLink.vis.inMultilink);
    if (!ml) throw Error('Unexpected.');
    setVisAttr(link, 'inMultilink', similarLink.vis.inMultilink);
    arr.push(_.cloneDeep(link));
    multilinkUpdateDirection(arr, ml);
  } else {
    const ml = createMultilink(
      availableId,
      groupId === source ? source : target, // source node of multilink should always be the group
      groupId === source ? target : source
    );
    const merging = [similarLink, link];
    merging.forEach(l => setVisAttr(l, 'inMultilink', ml.id));
    arr.push(_.cloneDeep(link));
    multilinkUpdateDirection(arr, ml);
    arr.push(ml);
    if (ml.source === ml.target)
      throw Error(
        `insertLinkWithMultilinkUpdate assertion failed: multilink ${
          ml.id
        } has source equal to target: ${JSON.stringify(ml)}`
      );
    return ml;
  }
  return null;
}

/**
 * Update the aggregated direction of multilink from links in given array
 * @param {Link[]} arr of links with links belonging to given multilink
 * @param {Link} ml multilink to update
 */
export function multilinkUpdateDirection(/* const */ arr, ml) {
  Vue.set(ml, 'direction', { nonDirectional: false, sourceToTarget: false, targetToSource: false });
  arr
    .filter(l => l.vis && l.vis.inMultilink === ml.id)
    .map(l => {
      if (!l.direction) {
        ml.direction.nonDirectional = true;
      } else {
        if (l.vis.newSource === ml.source) ml.direction.sourceToTarget = true;
        else ml.direction.targetToSource = true;
      }
    });
}

/**
 * Check if node is in aggregation, replaced by presentational node
 * @param {Node} node to check
 * @param {Number} id of presentational node
 */
export function isReplacedBy(node, id) {
  return !node.vis ? false : node.vis.replacedBy === id;
}

/**
 * Check if node is bundled, behind presentational node
 * @param {Node} node to check
 * @param {Number} id of presentational node
 */
export function isBundledIn(node, id) {
  return !node.vis ? false : node.vis.bundledIn === id;
}

export function isBundle(n) {
  return !n.vis ? false : n.vis.bundledNodes !== undefined && n.vis.bundledNodes.length > 0;
}

/**
 * Check if link is in multilink
 * @param {Link} link
 * @return {boolean}
 */
export function isInMultilink(link) {
  if (!link.vis) return false;
  return link.vis.inMultilink !== undefined;
}

/**
 * Check if link is in specific multilink
 * @param {Link} link
 * @param {number} mlId
 *
 * @return {boolean}
 */
export function isInMultilinkOfId(link, mlId) {
  if (!Number.isInteger(mlId)) throw Error('Invalid ID');

  if (!link.vis) return false;
  return link.vis.inMultilink === mlId;
}

/**
 * Check for vis.isMultilink attribute
 * @param {Link} link
 *
 * @return {boolean}
 */
export function isMultilink(link) {
  if (!link.vis) return false;
  return link.vis.isMultilink !== undefined;
}

/**
 * Check for vis.isArchived attribute
 * @param {Link} multilink
 *
 * @return {boolean}
 */
export function isMultilinkArchived(multilink) {
  if (!isMultilink(multilink)) return false;
  return multilink.vis.isArchived === true;
}

/**
 * Push new group node to state
 * @param {Object} state writes to state.nodes
 * @param {Node} node will be assigned type and new ID
 * @param {boolean} isUserDefined true to set this vis attr
 *
 * @return {number} ID assigned to the new node
 */
export function pushNodeToState(state, node, isUserDefined) {
  node.id = arrGetNewId(state.nodes);
  if (isUserDefined) {
    node.vis = node.vis || {};
    node.vis.isUserDefined = true;
  }
  state.nodes.push(_.cloneDeep(node));
  return node.id;
}

/**
 * Set vis attribute to object
 * @param {Node|Link} el
 * @param {string} attr vis attribute to set
 * @param {*} val value to set
 */
export function setVisAttr(el, attr, val) {
  el.vis = el.vis || {};
  Vue.set(el.vis, attr, _.cloneDeep(val));
}

/**
 * Set vis attributes to object, merging vis objects
 * @param {Node|Link} el
 * @param {NodeVis|LinkVis} visAttrs vis object to merge
 */
export function setVisAttrs(el, visAttrs) {
  el.vis = el.vis || {};
  Vue.set(el, 'vis', _.merge(el.vis, visAttrs));
}

/**
 * Merge visAttrs with each element given in arr
 * @param {Node[]|Link[]} arr
 * @param {LinkVis|NodeVis} visAttrs
 */
export function arraySetVisAttrs(arr, visAttrs) {
  arr.forEach(el => {
    el.vis = el.vis || {};
    Vue.set(el, 'vis', _.merge(el.vis, visAttrs));
  });
}

/**
 * Removes vis attr if set, then removes whole vis object if empty
 * @param {Node|Link} el
 * @param {string} attr attribute of vis object
 */
export function removeVisAttr(el, attr) {
  if (!el.vis) return;
  Vue.delete(el.vis, attr);
  if (Object.keys(el.vis).length === 0) Vue.delete(el, 'vis');
}

/**
 * Get new multilink instance
 * @param {number} newId
 * @param {number} source
 * @param {number} target
 */
export function createMultilink(newId, source, target) {
  // FIXME: when isUserDefined?
  return {
    id: newId,
    source: source,
    target: target,
    vis: { isMultilink: true }
  };
}

/**
 * Find last node which replaced given node
 * Nodes map is in form [{id: replacedById | undefined },..]
 * @param {Map} nodesMap created with makeNodesMap
 * @param {number} nodeId on which to start lookup
 *
 * @returns {number} node id
 */
export function findReplacementNode(nodesMap, nodeId) {
  if (!nodesMap.has(nodeId)) throw Error('Node not found in nodes map.');

  const replacedBy = nodesMap.get(nodeId);
  if (replacedBy === undefined) return nodeId;
  else return findReplacementNode(nodesMap, replacedBy);
}

/**
 * Check if link is transferred (connected to nodes inside bundle or to a group)
 * @param {Link} link to check
 * @param {Object} store vuex store instance to fetch node data
 */
export function isLinkTransferred(link, store) {
  if (link.vis && link.vis.forceTransferred) return true;

  const sourceNode = store.getters['data/getNodeById'](getLinkSource(link));
  const targetNode = store.getters['data/getNodeById'](getLinkTarget(link));

  switch (sourceNode.type) {
    case 'regular':
    case 'alias':
      switch (targetNode.type) {
        case 'regular':
        case 'alias':
          if (isMultilink(link)) {
            const singlelinks = store.getters['data/getMultilinkLinks'](link.id);
            for (const l of singlelinks) {
              if (l.vis && l.vis.forceTransferred) return true;
              if (
                isBundledIn(store.getters['data/getNodeById'](l.source), sourceNode.id) ||
                isBundledIn(store.getters['data/getNodeById'](l.target), targetNode.id)
              )
                return true;
            }
          } else {
            // check against original connection
            if (
              isBundledIn(store.getters['data/getNodeById'](link.source), sourceNode.id) ||
              isBundledIn(store.getters['data/getNodeById'](link.target), targetNode.id)
            )
              return true;
          }
          return false;
        case 'group':
          return true;
        default:
          throw Error(
            `Cannot compute isTransferred property: unknown link type of '${targetNode.type}'`
          );
      }
    case 'group':
      switch (targetNode.type) {
        case 'regular':
        case 'alias':
        case 'group':
          return true;
        default:
          throw Error(
            `Cannot compute isTransferred property: unknown link type of '${targetNode.type}'`
          );
      }
    default:
      throw Error(
        `Cannot compute isTransferred property: unknown link type of '${sourceNode.type}'`
      );
  }
}

/**
 * Create new link
 * @param {Object} newId to assign
 * @param {Number} source node id
 * @param {Number} target node id
 * @param {Object} [vis] vis attributes to set
 * @return created link instance
 */
export function createLink(newId, source, target, vis = null) {
  const l = { id: newId, target: target, source: source };
  if (vis !== null) {
    l.vis = l.vis || {};
    _.merge(l.vis, vis);
  }
  return l;
}

/**
 * Create Map object from state.nodes. Use to speed up lookups
 * @param {Object} state
 */
export function makeNodesMap(state) {
  return new Map(
    state.nodes.map(n => {
      if (n.vis) {
        if (n.vis.replacedBy !== undefined) return [n.id, n.vis.replacedBy];
        if (n.vis.bundledIn !== undefined) return [n.id, n.vis.bundledIn];
      }
      return [n.id, undefined];
    })
  );
}

/**
 * Validate dataset. Check that:
 * - every node/link has ID assigned
 * - node/link IDs are unique
 * - links are connected to nodes that exist
 * - links are unique between pair of nodes OR they are properly replaced by multilink
 *
 * @param {Object} dataset
 * @param {Node[]} dataset.nodes
 * @param {Link[]} dataset.links
 */
export function validateDataset(dataset) {
  if (!dataset) throw Error('Invalid input');

  const debug = false;
  debug && console.log('Validating dataset: ' + JSON.stringify(dataset));
  const data = _.cloneDeep(dataset);

  if (!data.nodes || !data.links) throw Error('Undefined nodes/links array');

  const nodes = new Map();
  data.nodes.forEach(n => {
    if (n.id === undefined) throw Error('Node must have ID assigned: ' + JSON.stringify(n));
    if (nodes.has(n.id)) throw Error('ID of node is not unique: ' + JSON.stringify(n));
    nodes.set(n.id, { links: [] });
  });
  debug && console.log('Nodes map: ' + JSON.stringify(Array.from(nodes)));

  const links = new Map();
  data.links.forEach(l => {
    if (l.id === undefined) throw Error('Link must have ID assigned: ' + JSON.stringify(l));
    if (links.has(l.id)) throw Error('ID of link is not unique: ' + JSON.stringify(l));
    links.set(l.id, undefined);

    const source = getLinkSource(l),
      target = getLinkTarget(l);
    if (!nodes.has(source) || !nodes.has(l.source) || !nodes.has(target) || !nodes.has(l.target))
      throw Error('Link is connected to nonexistent node');
    if (l.source === l.target)
      throw Error(`Link [id ${l.id}] has source [${l.source}] equal to target [${l.target}], 
        however, identity relation is not allowed.`);

    // fill nodes map with currently visible links conntected to them
    nodes.get(source).links.push(l);
    nodes.get(target).links.push(l);
  });
  debug && console.log('Nodes map after links: ' + JSON.stringify(Array.from(nodes)));
  debug && console.log('Links map: ' + JSON.stringify(Array.from(links)));

  // check link connections
  nodes.forEach(function(val, nodeId) {
    debug && console.log('Checking links for node: ' + nodeId + ' = ' + JSON.stringify(val.links));
    val.links.forEach(link => {
      debug && console.log('Link: ' + JSON.stringify(link));
      const source = getLinkSource(link);
      const target = getLinkTarget(link);

      const similarLinks = val.links.filter(l => isLinkSimilar(l, link) && l.id !== link.id);
      const multilinks = similarLinks.filter(l => isMultilink(l) && !isMultilinkArchived(l));
      const singlelinks = similarLinks.filter(l => !isMultilink(l));
      debug && console.log('Similar: ' + JSON.stringify(similarLinks));
      debug && console.log('   ML: ' + JSON.stringify(multilinks));
      debug && console.log('   other SL: ' + JSON.stringify(singlelinks));
      // there are similarly connected links except the link itself
      if (similarLinks.length > 0) {
        if (source === target) return;

        // if its multilink, check that there exists singlelinks which it replaced
        if (isMultilink(link)) {
          // the similar link is ml, so there should be at least one other singlelink
          if (!isMultilinkArchived(link) && singlelinks.length === 0)
            throw Error('Multilink ' + JSON.stringify(link) + ' was found without singlelinks.');
        } else {
          // archived multilink are not in the array
          if (multilinks.length !== 1)
            throw Error(
              'Nodes [' +
                source +
                ',' +
                target +
                '] have similar links ' +
                JSON.stringify(similarLinks) +
                ' with wrong multilinks set'
            );
          const ml = multilinks[0];

          if (link.vis.inMultilink !== ml.id)
            throw Error(
              'Link ' + link.id + ' is supposed to be in multilink ' + JSON.stringify(ml.id)
            );
        }
      } else {
        if (isMultilink(link) && !isMultilinkArchived(link))
          throw Error('Link ' + link.id + ' is not supposed to be multilink.');
      }
    });
  });
  console.log('Dataset validated.');
}

/**
 * Check if two number arrays contain the same elements
 * @param {number[]} a
 * @param {number[]} b
 */
export function arraysEqual(a, b) {
  return _.xor(a, b).length === 0;
}
