import { arraysEqual, arrGetNewId } from './helpers';

export const state = {
  /**
   * @type {Tag[]}
   * TODO: normalize
   */
  tags: [],
  tagData: { color: '', name: '' } // FIXME move to component state
};

/**
 * @typedef {Object} Tag
 * @property {number}   id
 * @property {string}   name
 * @property {number}   groupNodeId id of aggregation node created with tag
 * @property {number[]} assignedObjectsIds arr of object ids which have the tag
 * @property {boolean}  isVisualized if tag is visible in graph near the nodes
 * @property {string}   color in hexa
 * @property {string}   stroke in hexa
 * @property {'group'|'alias'|'bundle'|'manual'|'link'} type if created automatically by aggregation or manually by user
 */

export const mutations = {
  // FIXME: create documentation for all mutations and actions
  // create clear interface (payload attributes)

  /**
   *
   * @param {*} state
   * @param {Object} payload
   * @param {number[]} payload.arr array of object ids to tag
   * @param {number} payload.id id to be assigned
   * @param {string} payload.name name to be assigned
   * @param {string} payload.color color to be assigned
   */
  TAG_NODES(state, payload) {
    assignTags(state.tags, state.nodes, payload.arr, payload.id, payload.name, payload.color);
  },

  /**
   * Switches tag's visibility
   * @param {*} state
   * @param {string} name tag name
   */
  SWITCH_TAG_VISIBILITY(state, name) {
    let tag = state.tags.find(t => t.name === name);
    if (tag) tag.isVisualized = !tag.isVisualized;
  },

  /**
   * @param {*} state
   * @param {string} color tag color to be assigned
   */
  SET_TAG_COLOR(state, color) {
    state.tagData.color = color;
  },

  /**
   * Changes tag's color and stroke
   * @param {*} state
   * @param {Object} payload
   * @param {string} payload.name tag name
   * @param {string} payload.color tag color
   */
  CHANGE_TAG_COLOR_OF_TAG(state, payload) {
    let tag = state.tags.find(t => t.name === payload.name);
    tag.color = payload.color;
    if (tag.color === '') {
      tag.color = 'none';
      tag.stroke = '#000000';
    } else {
      tag.stroke = 'none';
    }
  },

  /**
   * @param {*} state
   * @param {string} name tag name to be assigned
   */
  SET_TAG_NAME(state, name) {
    state.tagData.name = name;
  }
};

export const actions = {
  tagNodes({ commit }, payload) {
    commit('TAG_NODES', {
      nodeIds: payload.nodesToTagIds,
      color: payload.color,
      name: payload.name
    });
  },

  switchTagVisibility({ commit }, name) {
    commit('SWITCH_TAG_VISIBILITY', name);
  },

  setTagColor({ commit }, color) {
    commit('SET_TAG_COLOR', color);
  },

  changeTagColorOfTag({ commit }, payload) {
    commit('CHANGE_TAG_COLOR_OF_TAG', { name: payload.name, color: payload.color });
  },

  setTagName({ commit }, name) {
    commit('SET_TAG_NAME', name);
  }
};

export const getters = {
  getTags: state => state.tags,
  tags: state => state.tags.map(tagId => state.tags[tagId]), // added, maybe add ids to the objects

  getTagsByNodeId: state => id => {
    return filterTagsByNodeId(state.tags, id);
  }
};

/**
 * Return tags which are either assigned or created on the given node
 * @param {Tag[]} tags
 * @param {number} id of node
 */
export function filterTagsByNodeId(tags, id) {
  let filteredTags = [];

  tags.forEach(tag => {
    if (tag.groupNodeId === id) filteredTags.push(tag);
    else if (tag.assignedObjectsIds.includes(id)) filteredTags.push(tag);
    // else skip
  });
  return filteredTags;
}

/**
 * Create new tag and push to tags array
 * @param {Tag[]}  tags from state
 * @param {Node[]} objects array of objects (Nodes, Links)
 * @property {number[]} assignedObjectsIds arr of node ids which have the tag
 * @param {number} tagId tag id to be assigned
 * @param {string}  name
 * @param {string}  color in hexa
 * @param {'group'|'alias'|'bundle'|'manual'|'link'} type if created automatically by aggregation or manually by user
 * @param {number} groupNodeId
 */
function createTag(tags, objects, assignedObjectsIds, tagId, name, color, type, groupNodeId) {
  if (type != 'bundle' && tags.find(t => t.name === name)) {
    throw Error('Tag with the same name already exists'); // FIXME: custom
  }

  let stroke = 'none';

  if (color === '') {
    color = 'none';
    stroke = '#000000';
  }

  if (name === '') name = 'group ' + tagId;

  /**
   * @type {Tag}
   */
  const tag = {
    id: tagId,
    name: name,
    isVisualized: true,
    color: color,
    stroke: stroke,
    type: type,
    groupNodeId: groupNodeId,
    assignedObjectsIds: assignedObjectsIds
  };
  tags.push(tag);
}

/**
 * Assigns tags to array of objects
 *
 * @param {Tag[]} tags from state
 * @param {Node[]} objects from state (Nodes, Links)
 * @param {number[]} arrIds array of object ids (Nodes, Links)
 * @param {number} tagId tag id to be assigned
 * @param {string} name tag name
 * @param {string} color tag color
 * @param {'group'|'alias'|'bundle'|'manual'} [type] aggregation type
 * @param {number} [groupNodeId] aggregation node Id
 */
export function assignTags(
  tags,
  objects,
  arrIds,
  tagId,
  name,
  color,
  type = 'manual',
  groupNodeId = null
) {
  /*arrIds.forEach(id => {
    //let obj = objects.find(o => o.id === id)
    //obj.vis.tags.push(tagId) // FIXME: dataset doesn't have vis.tags
  })*/

  createTag(tags, objects, arrIds, tagId, name, color, type, groupNodeId);
}

/**
 * If tag already exists on given objects, it updates it, else creates new tag
 * @param {Tag[]}  tags from state
 * @param {Node[]} objects from state (Nodes, Links)
 * @param {number[]} arrIds array of object ids
 * @param {'group'|'alias'|'bundle'} action aggregation type
 * @param {Node} groupNode aggregation node (we get obj ref. from data/AGGREGATE_NODES)
 * @param {string}  color in hexa
 */
export function updateTags(tags, objects, arrIds, action, groupNode, color) {
  for (let tag of tags) {
    // find equivalent tag
    // a tag is assigned after aggregation, but after dissagregating, it stays there, in case
    // the user wants to aggregate again using the tag
    if (arraysEqual(tag.assignedNodesIds, arrIds)) {
      if (action === 'bundle' && tag.groupNode.id === groupNode.id) {
        // FIXME invert the case
        // FIXME this if branch has no coverage
      } else {
        tag.groupNode = objects.find(n => n.id === groupNode.id); // FIXME groupNode is id, not a Node
        tag.name = tag.groupNode.label;
        tag.type = action; // if the tag was manually created before, and we are aggregating using the tag, save aggr. type
      }
      return;
      // if adding additional nodes to bundle, only update the tag, do not create new one
    } else if (action === 'bundle' && tag.groupNodeId === groupNode.id) {
      arrIds.forEach(id => {
        if (!tag.assignedObjectsIds.includes(id)) {
          //let obj = objects.find(o => o.id === id)
          //obj.vis.tags.push(id)  // FIXME: dataset doesn't have vis.tags
          tag.assignedObjectsIds.push(id);
        }
      });
      tag.name = groupNode.label;
      return;
    }
  }

  // for 'group' & 'alias': new group node and nodes within the group get new tag
  // for 'bundle': group node is not part of the bundle, it's only representative node
  let groupNodeId = action != 'bundle' ? groupNode.id : null;

  // console.log('tags', tags)
  // console.log('objects', objects)
  // console.log('arrIds', arrIds)
  // console.log('groupNode.label', groupNode.label)
  // console.log('action', action)
  // console.log('groupNodeId', groupNodeId)

  assignTags(
    tags,
    objects,
    arrIds,
    arrGetNewId(tags),
    groupNode.label,
    color,
    action, //'aggregation',
    groupNodeId
  );
}
